#define _POSIX_C_SOURCE 200809L

#include <lektor/database.h>
#include <lektor/macro.h>

#include <limits.h>
#include <stdio.h>
#include <string.h>

bool
database_config_set(sqlite3 *db, const char *section, const char *key, const char *value)
{
    static const char *SQL_STMT =
        "INSERT OR REPLACE INTO"
        " config (section, key, value)"
        " VALUES (?, ?, ?);\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    int code;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_BIND_TEXT(db, stmt, 3, value, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK && code != SQLITE_DONE) {
        fprintf(stderr, " ! database_config_set: Failed to insert or replace: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_text(sqlite3 *db, const char *section, const char *key, char *value, size_t len)
{
    static const char *SQL_STMT =
        "SELECT value"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    char *row;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);

    if (sqlite3_step(stmt) != SQLITE_ROW) {
        fprintf(stderr, " ! database_config_get_text: Failed to insert or replace: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    row = (char *) sqlite3_column_text(stmt, 0);
    strncpy(value, row, len);
    value[len - 1] = 0;
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_exists(sqlite3 *db, const char *section, const char *key)
{
    static const char *SQL_STMT =
        "SELECT value"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);

    if (sqlite3_step(stmt) != SQLITE_ROW) {
        fprintf(stderr, " ! database_config_exists: No rows: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_int(sqlite3 *db, const char *section, const char *key, int *value)
{
    static const char *SQL_STMT =
        "SELECT CAST(value AS INTEGER)"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);

    if (sqlite3_step(stmt) != SQLITE_ROW) {
        fprintf(stderr, " ! database_config_get_int: Failed to insert or replace: %s\n",
                sqlite3_errmsg(db));
        goto error;
    }

    *value = sqlite3_column_int(stmt, 1);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_queue(sqlite3 *db, const char *option, int value)
{
    static const char *SQL_STMT_TMP = "UPDATE queue_state SET %s = ? WHERE id = 42;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    int code;

    if (value < 0)
        value = 0;
    if (value > 100)
        value = 100;

    snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT - 1, SQL_STMT_TMP, option);
    SQL_STMT[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, value, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK || code != SQLITE_ROW || code != SQLITE_DONE)
        goto error;

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_queue_default(sqlite3 *db)
{
    static const char *SQL_DEFAULT =
        "UPDATE queue_state SET"
        "  volume = 100,"
        "  paused = 1,"
        "  random = 0,"
        "  repeat = 0,"
        "  single = 0,"
        "  consume = 0,"
        "  current = NULL,"
        "  duration = 0;";
    SQLITE_EXEC(db, SQL_DEFAULT, error);
    return true;
error:
    return false;
}

bool
database_get_config(sqlite3 *db, const char *option, int *value)
{
    static const char *SQL_STMT_TMP = "SELECT %s FROM queue_state WHERE id = 42;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT - 1, SQL_STMT_TMP, option);
    SQL_STMT[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_STEP_ROW(db, stmt, error);
    *value = sqlite3_column_int(stmt, 0);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}
